import uuid

from enum import Enum
from voluptuous import Schema, Invalid, MultipleInvalid


class Resource(object):
    DEFAULT_ID_LENGTH = 8

    def __init__(self, *args, **kwargs):
        if len(args):
            resource = args[0]
            if isinstance(resource, Resource):
                self.__id = resource.id
                self.__type = Resource.Type[resource.type]
            else:
                self.__attributes_from_dict(resource)
        else:
            self.__attributes_from_dict(kwargs)

    def __attributes_from_dict(self, attributes):
        schema = Schema({
            'id': str,
            'type': Resource.Type.validate
        })
        try:
            schema(attributes)
            self.__id = attributes.get('id', Resource._random_id())
            self.__type = Resource.Type[attributes.get('type', 'null')]
        except MultipleInvalid as invalid:
            raise InvalidResourceException(invalid)

    @property
    def id(self):
        return self.__id

    @property
    def type(self):
        return self.__type.name

    @staticmethod
    def _random_id():
        return uuid.uuid4().hex[:Resource.DEFAULT_ID_LENGTH]

    def __eq__(self, other):
        return isinstance(other, Resource) and \
               self.id == other.id and \
               self.type == other.type

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.id)

    def __hash__(self):
        return hash(self.__repr__())

    class Type(Enum):
        null = 0
        instance = 1

        @classmethod
        def validate(cls, value):
            try:
                cls[value]
            except KeyError:
                raise Invalid('The given resource type (%s) is not valid' % value)
            return True


class Instance(Resource):
    def __init__(self, *args, **kwargs):
        kwargs['type'] = Resource.Type.instance.name
        super(Instance, self).__init__(*args, **kwargs)


class ResourceGroup(object):
    def __init__(self, *args):
        self.__resources = set()
        if len(args):
            for argument in args:
                try:
                    self.__iadd__(argument)
                except InvalidResourceException:
                    for resource in argument:
                        self.__iadd__(resource)

    @property
    def resources(self):
        return list(self.__resources)

    def __add__(self, other):
        if isinstance(other, dict):
            return ResourceGroup(Resource(other), *self.resources)
        if isinstance(other, Resource):
            return ResourceGroup(other, *self.resources)
        if isinstance(other, ResourceGroup):
            resources = other.resources + self.resources
            return ResourceGroup(*resources)
        raise InvalidResourceException()

    def __iadd__(self, other):
        if isinstance(other, dict):
            self.__resources.add(Resource(other))
            return self
        if isinstance(other, Resource):
            self.__resources.add(other)
            return self
        if isinstance(other, ResourceGroup):
            self.__resources.union(other.resources)
            return self
        raise InvalidResourceException()

    def __iter__(self):
        return self.__resources.__iter__()

    def __eq__(self, other):
        return self.__resources == set(other.resources)


class InvalidResourceException(Exception):
    pass
