from unittest.case import TestCase

from voluptuous import Invalid

from orchestrator.model.resources import Resource, ResourceGroup, Instance, InvalidResourceException


class ResourceTest(TestCase):
    def test_resource_without_args(self):
        # when
        resource = Resource()

        # then
        self.assertIsInstance(resource.id, str)
        self.assertEqual(Resource.DEFAULT_ID_LENGTH, len(resource.id))
        self.assertEqual('null', resource.type)

    def test_resource_from_args(self):
        # given
        expected_id = 'dummy_resource'
        expected_type = 'null'

        # when
        result = Resource(id=expected_id, type=expected_type)

        # then
        self.assertEqual(expected_id, result.id)
        self.assertEqual(expected_type, result.type)

    def test_resource_from_dict(self):
        # given
        resource_dict = dict(id='dummy_resource', type='null')

        # when
        result = Resource(resource_dict)

        # then
        self.assertEqual(resource_dict['id'], result.id)
        self.assertEqual(resource_dict['type'], result.type)

    def test_resource_from_invalid_args(self):
        with self.assertRaises(InvalidResourceException):
            Resource(id='dummy_id', invalid_attribute=False)

    def test_resource_from_invalid_dict(self):
        # given
        invalid_attributes = dict(id='dummy_id', invalid_attribute=False)

        # when/then
        with self.assertRaises(InvalidResourceException):
            Resource(invalid_attributes)

    def test_resource_type_validate(self):
        self.assertTrue(Resource.Type.validate('instance'))

    def test_resource_type_invalid(self):
        with self.assertRaises(Invalid):
            Resource.Type.validate('nonexisting_resource_type')

    def test_resource_instance(self):
        # when
        instance = Instance()

        # then
        self.assertIsInstance(instance.id, str)
        self.assertEqual(Resource.DEFAULT_ID_LENGTH, len(instance.id))
        self.assertEqual(Resource.Type.instance.name, instance.type)

    def test_resource_equivalence(self):
        # given
        resource1 = Resource(id='some_resource', type='null')
        resource2 = Resource(id='some_resource', type='null')
        resource3 = Resource(id='some_other_resource', type='null')
        resource4 = Resource(id='some_resource', type='instance')

        # then
        self.assertEqual(resource1, resource1)
        self.assertEqual(resource1, resource2)
        self.assertNotEqual(resource1, resource3)
        self.assertNotEqual(resource1, resource4)


class ResourceGroupTest(TestCase):
    def test_resource_group_without_args(self):
        # when
        empty_group = ResourceGroup()

        # then
        self.assertEqual([], empty_group.resources)

    def test_resource_group_from_resources(self):
        # given
        resource1 = Resource(id='some_resource')
        resource2 = Resource(id='some_other_resource')

        # then
        self.assertEqual([resource1], ResourceGroup(resource1).resources)
        self.assertEqual([resource1], ResourceGroup([resource1]).resources)
        self.assertEqual([resource1], ResourceGroup([resource1, resource1]).resources)
        for resources in [[resource1, resource2], [resource2, resource1]]:
            self.assertEqual(2, len(ResourceGroup(*resources).resources))
            self.assertIn(resource1, ResourceGroup(*resources).resources)
            self.assertIn(resource2, ResourceGroup(*resources).resources)

    def test_resource_group_from_dicts(self):
        # given
        dict1 = dict(id='some_resource')
        dict2 = dict(id='some_other_resource')
        expected1 = Resource(dict1)
        expected2 = Resource(dict2)

        # then
        self.assertEqual([expected1], ResourceGroup(dict1).resources)
        self.assertEqual([expected1], ResourceGroup([dict1]).resources)
        self.assertEqual([expected1], ResourceGroup([dict1, dict1]).resources)
        for resources in [[expected1, expected2], [dict2, dict1]]:
            self.assertEqual(2, len(ResourceGroup(resources).resources))
            self.assertIn(expected1, ResourceGroup(resources).resources)
            self.assertIn(expected2, ResourceGroup(resources).resources)

    def test_resource_group_addition(self):
        # given
        resource1 = Resource()
        resource2 = Resource()
        group1 = ResourceGroup(resource1)
        group2 = ResourceGroup(resource2)

        # when
        result_group = group1 + group2

        # then
        self.assertEqual(2, len(result_group.resources))
        self.assertIn(resource1, result_group.resources)
        self.assertIn(resource2, result_group.resources)

    def test_resource_group_iteration(self):
        # when
        group = ResourceGroup(Resource(), Resource(), Resource())

        # then
        for index, resource in enumerate(group):
            self.assertEqual(group.resources[index], resource)

    def test_resource_group_equivalence(self):
        # given
        resource1 = Resource(id='some_resource', type='null')
        resource2 = Resource(id='some_resource', type='null')
        resource3 = Resource(id='some_other_resource', type='null')

        # then
        self.assertEqual(ResourceGroup(), ResourceGroup())
        self.assertEqual(ResourceGroup(resource1), ResourceGroup(resource1))
        self.assertEqual(ResourceGroup(resource1), ResourceGroup(resource2))
        self.assertNotEqual(ResourceGroup(resource1), ResourceGroup(resource3))
